@@ -50,10 +50,10 @@
 //!     let response = fetch_google_receipt_data(receipt, "<GOOGLE_KEY>").await?;
 //!
 //!     // debug or validate on your own with the data in the response
-//!     println!("Expiry data: {}", response.expiry_time);
+//!     println!("Expiry data: {:?}", response.expiry_time);
 //!
 //!     // or just simply validate the response
-//!     validate_google_subscription(&response)
+//!     validate_google_subscription(&response, chrono::Utc::now())
 //! }
 //! ```
 //!
@@ -66,7 +66,7 @@
 //!     // was this purchase made in the production or sandbox environment
 //!     println!("Environment: {}", response.environment.clone().unwrap());
 //!
-//!     Ok(validate_apple_subscription(&response))
+//!     Ok(validate_apple_subscription(&response, &receipt.transaction_id, chrono::Utc::now()))
 //! }
 //! ```
 
@@ -80,6 +80,8 @@
 #![deny(clippy::perf)]
 #![deny(clippy::nursery)]
 #![deny(clippy::match_like_matches_macro)]
+//TODO: remove once async_trait works with this again
+#![allow(clippy::no_effect_underscore_binding)]
 
 mod apple;
 mod google;
@@ -87,17 +89,18 @@ mod google;
 pub mod error;
 
 use async_trait::async_trait;
+use chrono::{DateTime, Utc};
 use error::Result;
 use serde::{Deserialize, Serialize};
 use yup_oauth2::ServiceAccountKey;
 
 pub use apple::{
-    fetch_apple_receipt_data, fetch_apple_receipt_data_with_urls, validate_apple_subscription,
-    AppleResponse, AppleUrls,
+    fetch_apple_receipt_data, fetch_apple_receipt_data_with_urls, validate_apple_package,
+    validate_apple_subscription, AppleResponse, AppleUrls,
 };
 pub use google::{
-    fetch_google_receipt_data, fetch_google_receipt_data_with_uri, validate_google_subscription,
-    GoogleResponse,
+    fetch_google_receipt_data, fetch_google_receipt_data_with_uri, validate_google_package,
+    validate_google_subscription, GoogleResponse, SkuType,
 };
 
 /// This is the platform on which the purchase that created the unity receipt was made.
@@ -144,13 +147,19 @@ impl UnityPurchaseReceipt {
 pub struct PurchaseResponse {
     /// Valid if true
     pub valid: bool,
+    /// Product identifier
+    pub product_id: Option<String>,
 }
 
 /// The base trait for implementing a validator. Mock Validators can be made for running local tests by implementing this trait.
 #[async_trait]
 pub trait Validator: Send + Sync {
     /// Called to perform the validation on whichever platform is described in the provided UnityPurchaseReceipt.
-    async fn validate(&self, receipt: &UnityPurchaseReceipt) -> Result<PurchaseResponse>;
+    async fn validate(
+        &self,
+        now: DateTime<Utc>,
+        receipt: &UnityPurchaseReceipt,
+    ) -> Result<PurchaseResponse>;
 }
 
 /// Trait which allows us to retrieve receipt data from an object's own secrets.
@@ -165,7 +174,7 @@ pub trait ReceiptDataFetcher {
     async fn fetch_google_receipt_data(
         &self,
         receipt: &UnityPurchaseReceipt,
-    ) -> Result<GoogleResponse>;
+    ) -> Result<(GoogleResponse, SkuType)>;
 }
 
 /// Convenience trait which combines `ReceiptDataFetcher` and `Validator` traits.
@@ -195,8 +204,9 @@ impl ReceiptValidator for UnityPurchaseValidator<'_> {}
 impl UnityPurchaseValidator<'_> {
     /// Stores Apple's shared secret required by their requestBody. See: <https://developer.apple.com/documentation/appstorereceipts/requestbody>
     #[allow(clippy::missing_const_for_fn)]
-    #[allow(clippy::must_use_candidate)]
+    #[must_use]
     pub fn set_apple_secret(self, secret: String) -> Self {
+        tracing::info!("Setting apple secret");
         let mut new = self;
         new.secret = Some(secret);
         new
@@ -223,8 +233,12 @@ impl UnityPurchaseValidator<'_> {
 
 #[async_trait]
 impl Validator for UnityPurchaseValidator<'_> {
-    async fn validate(&self, receipt: &UnityPurchaseReceipt) -> Result<PurchaseResponse> {
-        log::debug!(
+    async fn validate(
+        &self,
+        now: DateTime<Utc>,
+        receipt: &UnityPurchaseReceipt,
+    ) -> Result<PurchaseResponse> {
+        tracing::debug!(
             "store: {:?}, transaction_id: {}, payload: {}",
             receipt.store,
             &receipt.transaction_id,
@@ -239,43 +253,60 @@ impl Validator for UnityPurchaseValidator<'_> {
                     self.secret.as_ref(),
                 )
                 .await?;
+
                 if response.status == 0 {
-                    //apple returns latest_receipt_info if it is a renewable subscription
-                    match response.latest_receipt {
-                        Some(_) => Ok(validate_apple_subscription(&response)),
-                        None => unimplemented!("validate consumable"),
+                    if response.is_subscription(&receipt.transaction_id) {
+                        Ok(validate_apple_subscription(
+                            &response,
+                            &receipt.transaction_id,
+                            now,
+                        ))
+                    } else {
+                        Ok(validate_apple_package(&response, &receipt.transaction_id))
                     }
                 } else {
-                    Ok(PurchaseResponse { valid: false })
+                    Ok(PurchaseResponse {
+                        valid: false,
+                        product_id: response.get_product_id(&receipt.transaction_id),
+                    })
                 }
             }
             Platform::GooglePlay => {
                 //TODO: clean all of this up if async move evey makes its way to rust stable
-                if let Ok((Ok(response_future), Ok(sku_type))) =
-                    google::GooglePlayData::from(&receipt.payload).map(|data| {
-                        (
-                            data.get_uri().map(|uri| {
-                                fetch_google_receipt_data_with_uri(
-                                    self.service_account_key.as_ref(),
-                                    uri,
-                                )
-                            }),
-                            data.get_sku_details()
-                                .map(|sku_details| sku_details.sku_type),
-                        )
+                if let Ok((Ok(response_future), sku_type)) =
+                    google::GooglePlayData::from(&receipt.payload).and_then(|data| {
+                        data.get_sku_details().map(|sku_details| {
+                            let sku_type = sku_details.sku_type;
+                            (
+                                data.get_uri(&sku_type).map(|uri| {
+                                    fetch_google_receipt_data_with_uri(
+                                        self.service_account_key.as_ref(),
+                                        uri,
+                                        Some(data),
+                                    )
+                                }),
+                                sku_type,
+                            )
+                        })
                     })
                 {
                     if let Ok(response) = response_future.await {
-                        if sku_type == "subs" {
-                            validate_google_subscription(&response)
-                        } else {
-                            unimplemented!("validate consumable")
+                        match sku_type {
+                            google::SkuType::Subs => validate_google_subscription(&response, now),
+                            google::SkuType::Inapp => Ok(validate_google_package(&response)),
                         }
                     } else {
-                        Ok(PurchaseResponse { valid: false })
+                        Ok(PurchaseResponse {
+                            valid: false,
+                            product_id: None,
+                        })
                     }
                 } else {
-                    Ok(PurchaseResponse { valid: false })
+                    //TODO:
+                    Ok(PurchaseResponse {
+                        valid: false,
+                        product_id: None,
+                    })
                 }
             }
         }
@@ -294,12 +325,16 @@ impl ReceiptDataFetcher for UnityPurchaseValidator<'_> {
     async fn fetch_google_receipt_data(
         &self,
         receipt: &UnityPurchaseReceipt,
-    ) -> Result<GoogleResponse> {
+    ) -> Result<(GoogleResponse, SkuType)> {
+        let data = google::GooglePlayData::from(&receipt.payload)?;
+        let sku_type = data.get_sku_details()?.sku_type;
         fetch_google_receipt_data_with_uri(
             self.service_account_key.as_ref(),
-            google::GooglePlayData::from(&receipt.payload)?.get_uri()?,
+            data.get_uri(&sku_type)?,
+            Some(data),
         )
         .await
+        .map(|response| (response, sku_type))
     }
 }
 
@@ -307,7 +342,7 @@ impl ReceiptDataFetcher for UnityPurchaseValidator<'_> {
 mod tests {
     use super::*;
     use crate::{
-        apple::{AppleLatestReceipt, AppleResponse},
+        apple::{AppleInAppReceipt, AppleReceipt, AppleResponse},
         google::{validate_google_subscription, GoogleResponse},
     };
     use chrono::{Duration, Utc};
@@ -328,14 +363,288 @@ mod tests {
     #[tokio::test]
     #[serial]
     async fn test_sandbox_response() {
+        let expiry = (Utc::now() + Duration::days(1))
+            .timestamp_millis()
+            .to_string();
+        let apple_response = AppleResponse {
+            receipt: Some(AppleReceipt {
+                in_app: Some(vec![AppleInAppReceipt {
+                    product_id: Some("prod".to_string()),
+                    expires_date_ms: Some(expiry),
+                    transaction_id: Some("txn".to_string()),
+                    ..AppleInAppReceipt::default()
+                }]),
+                ..AppleReceipt::default()
+            }),
+            ..AppleResponse::default()
+        };
+
+        let _m1 = mock("POST", "/sb/verifyReceipt")
+            .with_status(200)
+            .with_body(&serde_json::to_string(&apple_response).unwrap())
+            .create();
+
+        let _m2 = mock("POST", "/verifyReceipt")
+            .with_status(200)
+            .with_body(r#"{"status": 21007}"#)
+            .create();
+
+        let url = &mockito::server_url();
+
+        let sandbox = format!("{}/sb", url);
+        let validator = new_for_test(url, &sandbox);
+
+        let response = validator
+            .validate(
+                Utc::now(),
+                &UnityPurchaseReceipt {
+                    transaction_id: "txn".to_string(),
+                    ..UnityPurchaseReceipt::default()
+                },
+            )
+            .await
+            .unwrap();
+
+        assert!(response.valid);
+        assert_eq!(response.product_id, Some("prod".to_string()));
+    }
+
+    #[tokio::test]
+    #[serial]
+    async fn test_validate_empty_transaction() {
+        let expiry = (Utc::now() + Duration::days(1))
+            .timestamp_millis()
+            .to_string();
+        let apple_response = AppleResponse {
+            receipt: Some(AppleReceipt {
+                in_app: Some(vec![
+                    AppleInAppReceipt {
+                        product_id: Some("prod".to_string()),
+                        expires_date_ms: Some(Utc::now().timestamp_millis().to_string()),
+                        transaction_id: Some("txn3".to_string()),
+                        ..AppleInAppReceipt::default()
+                    },
+                    AppleInAppReceipt {
+                        product_id: Some("prod".to_string()),
+                        expires_date_ms: Some(expiry),
+                        transaction_id: Some("txn1".to_string()),
+                        ..AppleInAppReceipt::default()
+                    },
+                    AppleInAppReceipt {
+                        product_id: Some("prod".to_string()),
+                        expires_date_ms: Some(
+                            (Utc::now() - Duration::days(1))
+                                .timestamp_millis()
+                                .to_string(),
+                        ),
+                        transaction_id: Some("txn2".to_string()),
+                        ..AppleInAppReceipt::default()
+                    },
+                ]),
+                ..AppleReceipt::default()
+            }),
+            ..AppleResponse::default()
+        };
+
+        let _m1 = mock("POST", "/sb/verifyReceipt")
+            .with_status(200)
+            .with_body(&serde_json::to_string(&apple_response).unwrap())
+            .create();
+
+        let _m2 = mock("POST", "/verifyReceipt")
+            .with_status(200)
+            .with_body(r#"{"status": 21007}"#)
+            .create();
+
+        let url = &mockito::server_url();
+
+        let sandbox = format!("{}/sb", url);
+        let validator = new_for_test(url, &sandbox);
+
+        let response = validator
+            .validate(
+                Utc::now(),
+                &UnityPurchaseReceipt {
+                    transaction_id: "".to_string(),
+                    ..UnityPurchaseReceipt::default()
+                },
+            )
+            .await
+            .unwrap();
+
+        assert!(response.valid);
+        assert_eq!(response.product_id, Some("prod".to_string()));
+    }
+
+    #[tokio::test]
+    #[serial]
+    async fn test_subscription_expired_fallback() {
+        let expiry = (Utc::now() + Duration::days(1))
+            .timestamp_millis()
+            .to_string();
+        let apple_response = AppleResponse {
+            receipt: Some(AppleReceipt {
+                in_app: Some(vec![
+                    AppleInAppReceipt {
+                        product_id: Some("prod".to_string()),
+                        expires_date_ms: Some(expiry),
+                        transaction_id: Some("txn1".to_string()),
+                        ..AppleInAppReceipt::default()
+                    },
+                    AppleInAppReceipt {
+                        product_id: Some("prod".to_string()),
+                        expires_date_ms: Some(
+                            (Utc::now() - Duration::days(1))
+                                .timestamp_millis()
+                                .to_string(),
+                        ),
+                        transaction_id: Some("txn2".to_string()),
+                        ..AppleInAppReceipt::default()
+                    },
+                ]),
+                ..AppleReceipt::default()
+            }),
+            ..AppleResponse::default()
+        };
+
+        let _m1 = mock("POST", "/sb/verifyReceipt")
+            .with_status(200)
+            .with_body(&serde_json::to_string(&apple_response).unwrap())
+            .create();
+
+        let _m2 = mock("POST", "/verifyReceipt")
+            .with_status(200)
+            .with_body(r#"{"status": 21007}"#)
+            .create();
+
+        let url = &mockito::server_url();
+
+        let sandbox = format!("{}/sb", url);
+        let validator = new_for_test(url, &sandbox);
+
+        // look up the expired receipt, should still be valid since txn1 is not expired
+        let response = validator
+            .validate(
+                Utc::now(),
+                &UnityPurchaseReceipt {
+                    transaction_id: "txn2".to_string(),
+                    ..UnityPurchaseReceipt::default()
+                },
+            )
+            .await
+            .unwrap();
+
+        assert!(response.valid);
+        assert_eq!(response.product_id, Some("prod".to_string()));
+    }
+
+    #[tokio::test]
+    #[serial]
+    async fn test_invalid_receipt() {
+        let now = Utc::now().timestamp_millis().to_string();
         let apple_response = AppleResponse {
             latest_receipt: Some(String::default()),
-            latest_receipt_info: Some(vec![AppleLatestReceipt {
-                expires_date_ms: (Utc::now() + Duration::days(1))
-                    .timestamp_millis()
-                    .to_string(),
-                ..AppleLatestReceipt::default()
-            }]),
+            receipt: Some(AppleReceipt {
+                in_app: Some(vec![AppleInAppReceipt {
+                    product_id: Some("prod".to_string()),
+                    expires_date_ms: Some(now),
+                    transaction_id: Some("txn".to_string()),
+                    ..AppleInAppReceipt::default()
+                }]),
+                ..AppleReceipt::default()
+            }),
+            ..AppleResponse::default()
+        };
+
+        let _m = mock("POST", "/verifyReceipt")
+            .with_status(200)
+            .with_body(&serde_json::to_string(&apple_response).unwrap())
+            .create();
+
+        let url = &mockito::server_url();
+
+        let sandbox = format!("{}/sb", url);
+        let validator = new_for_test(url, &sandbox);
+
+        assert!(
+            !validator
+                .validate(
+                    Utc::now(),
+                    &UnityPurchaseReceipt {
+                        transaction_id: "txn".to_string(),
+                        ..UnityPurchaseReceipt::default()
+                    }
+                )
+                .await
+                .unwrap()
+                .valid
+        );
+    }
+
+    #[tokio::test]
+    #[serial]
+    async fn test_apple_purchase() {
+        let apple_response = AppleResponse {
+            receipt: Some(AppleReceipt {
+                in_app: Some(vec![
+                    AppleInAppReceipt {
+                        product_id: Some("prod".to_string()),
+                        transaction_id: Some("not_txn".to_string()),
+                        ..AppleInAppReceipt::default()
+                    },
+                    AppleInAppReceipt {
+                        product_id: Some("prod".to_string()),
+                        transaction_id: Some("txn".to_string()),
+                        ..AppleInAppReceipt::default()
+                    },
+                ]),
+                ..AppleReceipt::default()
+            }),
+            ..AppleResponse::default()
+        };
+
+        let _m1 = mock("POST", "/sb/verifyReceipt")
+            .with_status(200)
+            .with_body(&serde_json::to_string(&apple_response).unwrap())
+            .create();
+
+        let _m2 = mock("POST", "/verifyReceipt")
+            .with_status(200)
+            .with_body(r#"{"status": 21007}"#)
+            .create();
+
+        let url = &mockito::server_url();
+
+        let sandbox = format!("{}/sb", url);
+        let validator = new_for_test(url, &sandbox);
+
+        let response = validator
+            .validate(
+                Utc::now(),
+                &UnityPurchaseReceipt {
+                    transaction_id: "txn".to_string(),
+                    ..UnityPurchaseReceipt::default()
+                },
+            )
+            .await
+            .unwrap();
+
+        assert!(response.valid);
+        assert_eq!(response.product_id, Some("prod".to_string()));
+    }
+
+    #[tokio::test]
+    #[serial]
+    async fn test_apple_purchase_failed_missing_txn() {
+        let apple_response = AppleResponse {
+            receipt: Some(AppleReceipt {
+                in_app: Some(vec![AppleInAppReceipt {
+                    product_id: Some("prod".to_string()),
+                    transaction_id: Some("not_txn".to_string()),
+                    ..AppleInAppReceipt::default()
+                }]),
+                ..AppleReceipt::default()
+            }),
             ..AppleResponse::default()
         };
 
@@ -354,86 +663,15 @@ mod tests {
         let sandbox = format!("{}/sb", url);
         let validator = new_for_test(url, &sandbox);
 
-        assert!(
-            validator
-                .validate(&UnityPurchaseReceipt::default())
-                .await
-                .unwrap()
-                .valid
-        );
-    }
-
-    #[tokio::test]
-    #[serial]
-    async fn test_invalid_receipt() {
-        let apple_response = AppleResponse {
-            latest_receipt: Some(String::default()),
-            latest_receipt_info: Some(vec![AppleLatestReceipt {
-                expires_date_ms: Utc::now().timestamp_millis().to_string(),
-                ..AppleLatestReceipt::default()
-            }]),
-            ..AppleResponse::default()
-        };
-
-        let _m = mock("POST", "/verifyReceipt")
-            .with_status(200)
-            .with_body(&serde_json::to_string(&apple_response).unwrap())
-            .create();
-
-        let url = &mockito::server_url();
-
-        let sandbox = format!("{}/sb", url);
-        let validator = new_for_test(url, &sandbox);
-
         assert!(
             !validator
-                .validate(&UnityPurchaseReceipt::default())
-                .await
-                .unwrap()
-                .valid
-        );
-    }
-
-    #[tokio::test]
-    #[serial]
-    async fn test_most_recent_receipt() {
-        let now = Utc::now().timestamp_millis();
-        let day = Duration::days(1).num_milliseconds();
-
-        let latest_receipt_info = vec![
-            AppleLatestReceipt {
-                expires_date_ms: now.to_string(),
-                ..AppleLatestReceipt::default()
-            },
-            AppleLatestReceipt {
-                expires_date_ms: (now + day).to_string(),
-                ..AppleLatestReceipt::default()
-            },
-            AppleLatestReceipt {
-                expires_date_ms: (now - day).to_string(),
-                ..AppleLatestReceipt::default()
-            },
-        ];
-
-        let apple_response = AppleResponse {
-            latest_receipt: Some(String::default()),
-            latest_receipt_info: Some(latest_receipt_info),
-            ..AppleResponse::default()
-        };
-
-        let _m = mock("POST", "/verifyReceipt")
-            .with_status(200)
-            .with_body(&serde_json::to_string(&apple_response).unwrap())
-            .create();
-
-        let url = &mockito::server_url();
-
-        let sandbox = format!("{}/sb", url);
-        let validator = new_for_test(url, &sandbox);
-
-        assert!(
-            validator
-                .validate(&UnityPurchaseReceipt::default())
+                .validate(
+                    Utc::now(),
+                    &UnityPurchaseReceipt {
+                        transaction_id: "txn".to_string(),
+                        ..UnityPurchaseReceipt::default()
+                    }
+                )
                 .await
                 .unwrap()
                 .valid
@@ -455,7 +693,7 @@ mod tests {
 
         assert!(
             !validator
-                .validate(&UnityPurchaseReceipt::default())
+                .validate(Utc::now(), &UnityPurchaseReceipt::default())
                 .await
                 .unwrap()
                 .valid
@@ -466,7 +704,7 @@ mod tests {
     #[serial]
     async fn test_google_fail() {
         let google_response = GoogleResponse {
-            expiry_time: Utc::now().timestamp_millis().to_string(),
+            expiry_time: Some(Utc::now().timestamp_millis().to_string()),
             ..GoogleResponse::default()
         };
 
@@ -479,9 +717,10 @@ mod tests {
 
         assert!(
             !validate_google_subscription(
-                &google::fetch_google_receipt_data_with_uri(None, url.clone())
+                &google::fetch_google_receipt_data_with_uri(None, url.clone(), None,)
                     .await
-                    .unwrap()
+                    .unwrap(),
+                Utc::now()
             )
             .unwrap()
             .valid
@@ -508,9 +747,12 @@ mod tests {
     #[serial]
     async fn test_google() {
         let google_response = GoogleResponse {
-            expiry_time: (Utc::now() + Duration::days(1))
-                .timestamp_millis()
-                .to_string(),
+            product_id: Some("prod".to_string()),
+            expiry_time: Some(
+                (Utc::now() + Duration::days(1))
+                    .timestamp_millis()
+                    .to_string(),
+            ),
             ..GoogleResponse::default()
         };
         let _m = mock("GET", "/test")
@@ -520,14 +762,41 @@ mod tests {
 
         let url = &mockito::server_url();
 
-        assert!(
-            validate_google_subscription(
-                &google::fetch_google_receipt_data_with_uri(None, url.clone())
-                    .await
-                    .unwrap()
-            )
-            .unwrap()
-            .valid
+        let response = validate_google_subscription(
+            &google::fetch_google_receipt_data_with_uri(None, url.clone(), None)
+                .await
+                .unwrap(),
+            Utc::now(),
+        )
+        .unwrap();
+
+        assert!(response.valid);
+        assert_eq!(response.product_id, Some("prod".to_string()));
+    }
+
+    #[tokio::test]
+    #[serial]
+    async fn test_google_purchase() {
+        let google_response = GoogleResponse {
+            purchase_state: Some(0),
+            product_id: Some("prod".to_string()),
+            ..GoogleResponse::default()
+        };
+
+        let _m = mock("GET", "/test")
+            .with_status(200)
+            .with_body(&serde_json::to_string(&google_response).unwrap())
+            .create();
+
+        let url = &mockito::server_url();
+
+        let response = validate_google_package(
+            &google::fetch_google_receipt_data_with_uri(None, url.clone(), None)
+                .await
+                .unwrap(),
         );
+
+        assert!(response.valid);
+        assert_eq!(response.product_id, Some("prod".to_string()));
     }
 }
